
from .ingestion.time_series_ingestion import *
from .utils.ts_utils import *

from .src.filtering import *
from .src.multiproc_filtering import *

from .src.filtering_with_dev import *
from .src.multiproc_filtering_w_dev import *

from .dataviz.interactive_viz import *



"""
This module contains filtering functions with deviation from median
"""

import pandas as pd
import numpy as np


def hampel_filter_with_dev_df(df: pd.DataFrame, vals_col: str, time_col=None, win_size=30, num_dev=3,
                              center_win=True) -> pd.DataFrame:
    """
    This function takes in dataframe containing time series of values, applies Hampel filter on
    these values, and returns dataframe consisting of original values columns along with
    the Hampel filtered data, outlier values, boolean flags where outliers found, values for lower
    deviation from median, values for upper deviation from median.

    Parameters
    ----------
    df: pd.DataFrame
        data from containing time series that needs to be Hampel filtered
    vals_col: str
        Single column name that contains values that need to be filtered.
    time_col: str
        Name of column that contains dates or timestamps
    win_size: int
        Size of sliding window for filtering.  Essentially the number of time steps to be considered when filtering.
    num_dev: int
        Number of standard deviations to consider when detecting values that would be considered outliers.
    center_win: Boolean
        Boolean value that determines whether the window is centered about the point being filtered?  Default=True.
        If False, point is at the leading edge (i.e. right side) of window  calculation.

    Returns
    -------
    Function returns a full dataframe consisting of original values columns along with
    the Hampel filtered data, outlier values and boolean flags where outliers found.
    """

    if (time_col != None):
        if (time_col not in list(df.columns)):
            raise Exception("Timestamp column '{}' is missing!".format(time_col))
        elif (time_col in list(df.columns)):
            if (not np.issubdtype(df[time_col].dtype, np.datetime64)):
                if (not np.issubdtype(pd.to_datetime(df[time_col]).dtype, np.datetime64)):
                    raise Exception("Timestamp column '{}' is not np.datetime64".format(time_col))
                else:
                    df[time_col] = pd.to_datetime(df[time_col])
                    drop_cols = set(df.columns) - set([time_col, vals_col])
                    # Not really filtered at this point. Just naming appropriately ahead of time.
                    orig_vals = df.sort_values(time_col, ascending=True).set_index(time_col).copy()
                    filtered = orig_vals.drop(columns=drop_cols).copy()
            else:
                df[time_col] = pd.to_datetime(df[time_col])
                drop_cols = set(df.columns) - set([time_col, vals_col])
                # Not really filtered at this point. Just naming appropriately ahead of time.
                orig_vals = df.sort_values(time_col, ascending=True).set_index(time_col).copy()
                filtered = orig_vals.drop(columns=drop_cols).copy()

    elif (time_col == None):
        if (not isinstance(df.index, pd.DatetimeIndex)):
            raise Exception("DataFrame index is not pd.DatetimeIndex")
        else:
            df.sort_index(inplace=True)
            drop_cols = set(df.columns) - set([vals_col])
            orig_vals = df.copy()
            filtered = orig_vals.drop(columns=drop_cols).copy()

    # Scale factor for estimating standard deviation based upon median value
    L = 1.4826

    # Calculate rolling median for the series
    rolling_median = filtered.rolling(window=int(win_size), center=center_win, min_periods=1).median()

    # Define a lambda function to apply to the series to calculate Median Absolute Deviation
    MAD = lambda x: np.median(np.abs(x - np.median(x)))

    # Calculate rolling MAD series
    rolling_MAD = filtered.rolling(window=(win_size), center=center_win, min_periods=1).apply(MAD)

    # Calculate threshold level for filtering based upon the number of standard deviation and
    # constant scaling factor L.
    threshold = int(num_dev) * L * rolling_MAD

    # Difference between original values and rolling median
    # Again, "filtered" not yet filtered at this point.
    difference = np.abs(filtered - rolling_median)

    median_minus_threshold = rolling_median - threshold
    median_minus_threshold.rename(columns={vals_col: 'LOWER_DEV'}, inplace=True)
    median_plus_threshold = rolling_median + threshold
    median_plus_threshold.rename(columns={vals_col: 'UPPER_DEV'}, inplace=True)

    '''
    # TODO: Look at logic here to possibly not mark as an outlier if threshold value
    is 0.0
    '''

    # Flag outliers
    outlier_idx = difference > threshold

    # Now it's filtered.  This should replace original values with filtered values from the rolling_median
    # dataframe where outliers were found.
    filtered[outlier_idx] = rolling_median[outlier_idx]
    filtered.rename(columns={vals_col: 'FLTRD_VAL'}, inplace=True)

    # Capture outliers column
    outliers = orig_vals[outlier_idx].rename(columns={vals_col: 'OUTLIER_VAL'}).drop(columns=drop_cols)
    # Capture outlier IS_OUTLIER column
    outlier_idx.rename(columns={vals_col: 'IS_OUTLIER'}, inplace=True)

    # The following returns a full dataframe consisting of original values columns
    # along with the Hampel filtered data, outlier values and boolean flags where outliers found.
    return pd.concat([orig_vals, filtered, outliers, outlier_idx, median_minus_threshold, median_plus_threshold], axis=1)


def hampel_filter_dev_list_of_ts(ts_list: list, vals_col: str, time_col=None,
                             win_size: int = 30, num_dev: int = 3, center_win: bool = True):
    """
    Function returns a list of filtered dataframes, each consisting of original columns along with
    the Hampel filtered data, outlier values and boolean flags where outliers found.

    Parameters
    ----------
    ts_list: list
        List of time-series dataframes
    vals_col: str
        Single column name that contains values that need to be filtered.
    time_col: str
        Name of column that contains dates or timestamps
    win_size: int
        Size of sliding window for filtering.  Essentially the number of time steps to be considered when filtering.
    num_dev: int
        Number of standard deviations to consider when detecting values that would be considered outliers.
    center_win: Boolean
        Boolean value that determines whether the window is centered about the point being filtered?  Default=True.
        If False, point is at the leading edge (i.e. right side) of window  calculation.

    Returns
    -------
    Returns a list of filtered dataframes, each consisting of original columns along with
    the Hampel filtered data, outlier values and boolean flags where outliers found.
    """
    filtered_ts = []

    for ts in ts_list:
        filtered_ts.append(hampel_filter_with_dev_df(ts, vals_col, time_col, win_size, num_dev, center_win))
    return filtered_ts



"""
This module contains filtering function that leverage multiprocessor capability
"""

import multiprocessing as mp
import pandas as pd

__all__ = ["hampel_mp"]


def hampel_mp(g_list: list, vals_col: str, time_col=None, win_size=30, num_dev=3, center_win=True):
    """

    Parameters
    ----------
    g_list: list(pd.DataFrame)
        list of time series dataframes usually generated from gen_list_of_frames() function, but can also be
        individually compiled list of time series dataframes as long as columns are consistently named in each one.
    vals_col: str
        Name of column containing the original values to be processed.
    time_col: str
        Name of column containing datetime for time series.  This is optional if dataframe index consists of
        DatetimeIndex.  If time_col contains string representation of date, it will be converted to datetime64.
    win_size: int
        Value of width (time periods) of sliding window.  If time periods are in days, this is the width in number
        of days.  If time periods are in hours, this is the width in hours.
    num_dev: int
        Number of standard deviations used in calculating threshold used to determine an outlier.
    center_win: boolean
        Boolean to indicate if point considered is at center of window or leading edge of window.  If False,
        calculations are similar to moving median with point to be considered on right side of window and points in
        past used for MAD calculation.

    Returns
    -------
    Function returns a list of dataframes,  with each dataframe consisting of original values columns along with
    the Hampel filtered data, outlier values and boolean flags where outliers found.

    """

    # The code for the Hampel filter needed to be added to a separate file for
    # this multiprocessing call to work.  HampelFilter.py needs to be in the local directory.
    from HampelFiltering import hampel_filter_df

    __name__ = '__main__'

    global results
    results = []
    if __name__ == '__main__':
        # Step 1: Init multiprocessing.Pool()
        pool = mp.Pool(mp.cpu_count())

        # Step 2: Define callback function to collect the output in 'results' to take
        # advantage of asynchronous calls.
        def collect_result(result):
            # global results
            results.append(result)

        # Step 3: Use loop to parallelize. This asynchronous call uses the collect_result
        # function defined earlier and adds results to results list.
        # This uses sliding window size of window_size and num_dev standard deviations for outlier detection.
        for row in g_list:
            pool.apply_async(hampel_filter_df,
                             args=(row, vals_col, time_col, win_size,num_dev, center_win),
                             callback=collect_result)

        # Step 4: Close Pool and let all the processes complete
        pool.close()
        pool.join()  # postpones the execution of next line of code until all processes in the queue are done.

        return results





import importlib
import inspect

from .filtering import *
from .multiproc_filtering import *
from .filtering_with_dev import *
from .multiproc_filtering_w_dev import *

code_hampel_with_dev = 'HampelWithDev.py'
code_out1 = open(code_hampel_with_dev, 'w')

print("Writing local code files needed for multiprocessing")

moda = importlib.import_module('pyhampel.src.HampelWithDev')
src_code = inspect.getsource(moda)
code_out1.write(src_code)
code_out1.close()

# second file
code_hampel = 'HampelFiltering.py'
code_out2 = open(code_hampel, 'w')

# print("Writing local code file needed for multiprocessing")

modb = importlib.import_module('pyhampel.src.HampelFiltering')
src_code2 = inspect.getsource(modb)
code_out2.write(src_code2)
code_out2.close()
